/* wagon-ESP-OSC
 *  un wagon radio-guidé
 *  plateforme ESP8266
 *  commande moteur Olimex BB-L298
 *  protocole réseau OSC
 *  SNHack gepeto@du-libre.org
 *  Codes sous licences GPLV3
 */

/* projet LOCO-DOM
 *  IP du réseau 192.168.0.x/24
 *  les loco sont de x= 81 a 91
 *  ports 9081 a 9091
 * Connexions
 * Olimex ESP8266-EVB |con3<=>BBL-298
 * |1 | 1| +5 |
 * |2 | 8| GND |
 * Olimex UEXT <=> BB-298
 * | 1  | 3.3V   |     | nc  |
 * | 2  | GND    |     | console debug GND |
 * | 3  | gpio 1 | TXD | console debug rx |
 * | 4  | gpio 3 | RXD | console debug tx |
 * | 5  | gpio 4 |   6 | nc |
 * | 6  | gpio 2 |   7 | nc |
 * | 7  | gpio 12 |  4 | in 2 |
 * | 8  | gpio 13 |  3 | in 1 |
 * | 9  | gpio 14 |  2 | enable A |
 * | 10 | gpio 15 |  5 | rfid tx |
 * + lecteur rfid
 */

// WIFI
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <Ticker.h> // gestion des timer

//MDNSResponder mdns;
// OSC
#include <OSCBundle.h>
#include <OSCMessage.h>
#include <OSCBoards.h>
//port numbers for OSC
const unsigned int inPort = 9081;
//destination IP pour la trace OSC
IPAddress MonitIP(192, 168, 0, 123); // IP et port du moniteur
const unsigned int MonitPort = 9981;
WiFiUDP udpin;
WiFiUDP udpout;
Ticker zou; // OSC envoi position
Ticker init_rfid; // init port rfid
// serial
#include <EspSoftSerialRx.h>
EspSoftSerialRx rfid;
const char* ssid     = "LOCOMO";
const char* password = "0123456789";
char nomhost[20]  = "LOCO-1";
// DHCP donc pas besoin
//IP
//IPAddress (192,168,IPC,IPD);
int IPC = 0;
int IPD = 81;
String nom = "LOCO-1";
char* OSCVar = "/LOCO-1";

// connect motor controller pins to Arduino digital pins
// ESP moteur traction
int enA = 12;
int in1 = 13;
int in2 = 14;
int Vitesse_actuelle = 0 ; // vitesse actuelle, <0 = arriere
uint32_t temps_boucle_total = 0 ; // en milli seconde
uint32_t temps_a_faire = 0;
uint32_t temps_depart = 0;
int rampe = 50;

const int nbr_de_puces = 20;
int no_puce_actuelle = 98;
int puce_a_trouver = 98;
int puce_derniere_lue = 98;
// tableau des puces rfid 20x13
char *RFID[nbr_de_puces * 13] = {
  "6D0070007964", // puce 0
  "010611B52F8C", // 1
  "0102BBAEF1E7", // 2
  "010445C652D4", // 3
  "010445D771E6", // 4
  "010445D8138B", // 5
  "0106116C7309", // 6
  "010611B118BF", // 7
  "05004E236C04", // 8
  "05004E23F79F", // 9
  "05004D1DE6B3", // 10
  "05004D0FEBAC", // 11
  "05004E23630B", // 12
  "1123456ab1",
  "1123456ab2",
  "1123456ab3",
  "1123456ab4",
  "1123456ab5",
  "1123456ab6",
  "1123456ab7",
  "1123456ab8",
};


static int32_t sequencenumber = 0;
int commande = 0; // no de la commande en cours

void setup()
{
  Serial.begin(115200);
  // set all the motor control pins to outputs
  pinMode(enA, OUTPUT);
  pinMode(in1, OUTPUT);
  pinMode(in2, OUTPUT);
  // serial RFID
  rfid.begin(9600, 2); // sortie GPIO2
  //setup ethernet part
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  //  DHCP donc pas fixé ici
  //WiFi.config(IPAddress(192, 168, IPC, IPD), IPAddress(192, 168, IPC, IPD), IPAddress(255, 255, 255, 0));
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println(ssid);
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  Serial.println("host: ");
  Serial.println(WiFi.hostname());
  WiFi.hostname(nomhost);
  Serial.println(WiFi.hostname());
  Serial.println("Starting UDP");
  udpin.begin(inPort);
  Serial.print("Local port: ");
  Serial.println(udpin.localPort());
  /* // pas de mDNS ca pas de DHCP sur l'AP?
   * if (mdns.begin(nomhost, WiFi.softAPIP())) {
    Serial.println("MDNS responder started");
    Serial.print("You can now connect to http://");
    Serial.print(nomhost);
    Serial.println(".local");
    // Add service to MDNS-SD
    //MDNS.addService("http", "tcp", 80);
  } else {
    Serial.println("pb mdns");
  }
  */
  // placer jesuisla en boucle timer 1s
  delay(100);
  zou.attach(5, jesuisla);
  // init_rfid.attach(10, rfid_port_init);

  // Avant;
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
  Vitesse_moteur(1000);
  //analogWrite(enA, 1000);
}
void test() { // envoie OSC /msg if position vitesse
  Serial.println("jesuisla");
}
void rfid_port_init() { // re initialisation du port rfid les 10s
  Serial.println("init rfid");
  rfid.service();
}
void jesuisla() { // envoie OSC /msg if position vitesse
  Serial.println("jesuisla");
  OSCMessage zouzou(OSCVar);
  zouzou.add((int32_t)no_puce_actuelle);
  zouzou.add((int32_t)Vitesse_actuelle);
  udpout.beginPacket(MonitIP, MonitPort);
  zouzou.send(udpout); // envoi au WIFIudp
  udpout.endPacket(); // mark the end of the OSC Packet
  zouzou.empty(); // free space occupied by
  OSCMessage temps("/TIME");
  temps.add((int32_t)millis());
  temps.send(udpout);
  udpout.endPacket();
  temps.empty();
}
void Arriere() {
  digitalWrite(in1, LOW);
  digitalWrite(in2, HIGH);
}
void Avant() {
  digitalWrite(in1, HIGH);
  digitalWrite(in2, LOW);
}
int Lit_rfid() {
  // mise en forme de la lecture rfid
  //retourne le no_puce

  //Serial.println("attente rfid");
  int i = 0;

  char tagValue[12];
  for (i = 0; i <= 11; i++) {
    unsigned char val;
    if (rfid.read(val)) {
      tagValue[i] = '\0'; // Null
      // on saute avec le start 0x02 et stop bit 0x03
      if ((val == 0x0D) || (val == 0x0A) || (val == 0x03) || (val == 0x02)) {
        break;
      }
      //      Serial.print(i);
      //      Serial.print(" ");
      //      Serial.print((char)val);
      //      Serial.print(" ");
      //    Serial.println(val, HEX );
      tagValue[i] = (char)val;

      if ((i == 11) ) { // 11 digit read is complete
        tagValue[12] = '\0'; // Null
        //        Serial.print("Tag read: ");
        Serial.println(tagValue);
        // Search the tag database for this particular tag
        int tagId = CherchePuce( tagValue );
        // Only fire the strike plate if this tag was found in the database
        if ( tagId != 99 )
        {
          Serial.print("rfid puce ");
          Serial.println(tagId);
          Message(tagId);
          return (tagId);
        } else {
          Serial.println("rfid puce ?");
        }
      }
    }
  }
}

int CherchePuce( char tag[12] ) {
  for (int puce = 0; puce < nbr_de_puces; puce++) {
    // Check if the tag value matches this row in the tag database
    Serial.println(RFID[puce]);
    if (strcmp(tag, RFID[puce]) == 0)
      return (puce);
  }
  // Si pas trouvé on retourne 99
  return (99);
}


bool Message(int bla) { // envoie OSC /msg i position
  OSCMessage envoi(OSCVar);
  envoi.add(bla);
  //envoi.add("/sequencenumber").add(sequencenumber++);
  udpout.beginPacket(MonitIP, MonitPort);
  envoi.send(udpout); // envoi au WIFIudp
  udpout.endPacket(); // mark the end of the OSC Packet
  envoi.empty(); // free space occupied by
  Serial.println("Message envoi");
  Serial.println(bla);
}

bool Aller(int no_puce, int vitesse) {
  // aller jusqu'a la puce no_puce a vitesse
  Serial.println("Aller");
  int tag;
  if (no_puce == no_puce_actuelle) {
    Message(no_puce);
    return true;
  }
  Vitesse_moteur(vitesse);
  while (tag = Lit_rfid() != no_puce) { // roule tant que puce pas bonne
    Message(tag);
    delay(100);
  }
  no_puce_actuelle = no_puce;
  return (true); // ? retourner le temps .. ///?///
}

int Vitesse_moteur(int moteur) {
  int i = 0;
  if (moteur > 1024) // filtre le max
    moteur = 1024;

  if (moteur < -1024)
    moteur = -1024;

  Serial.println(moteur);
  Serial.println(Vitesse_actuelle);
  if (Vitesse_actuelle > moteur) {
    for (i = Vitesse_actuelle; i >= moteur; i = i - 20) {
      if ( i >= 0 ) {
        //Avant;
        digitalWrite(in1, HIGH);
        digitalWrite(in2, LOW);
      } else {
        //Arriere;
        digitalWrite(in1, LOW);
        digitalWrite(in2, HIGH);
      }
      analogWrite(enA, abs(i));
      delay(50); // rampe
    }
  } else {
    for ( i = Vitesse_actuelle; i <= moteur; i = i + 20) {
      if ( i >= 0 ) {
        //Avant;
        digitalWrite(in1, HIGH);
        digitalWrite(in2, LOW);
      } else {
        //Arriere;
        digitalWrite(in1, LOW);
        digitalWrite(in2, HIGH);
      }
      analogWrite(enA, abs(i));
      delay(rampe); // rampe
    }
  }
  Vitesse_actuelle = moteur;
  analogWrite(enA, abs(moteur));
  no_puce_actuelle = 97;
}
// COMMANDES
void ArretMoteur(OSCMessage &msg) {
  commande = 0;
  Serial.println("ArretMoteur /stop");
  analogWrite(enA, 0);
}
void MoteurControl(OSCMessage &msg) {
  // met le moteur à la vitesse reçue
  commande = 1;
  Serial.println("/moteur");
  int moteur_lu = msg.getFloat(0);
  Vitesse_moteur(moteur_lu);
}
void Initialisation(OSCMessage &msg) {       // A FINIR
  // la loco fait un tour complet pour connaitre le temps nécessaire
  // et vérifier la lecture des puces RFID , voir faire l'init du tableau?
  // elle ne connait au départ que la puce 0, position 0
  commande = 2;
  //moteur 0
  Vitesse_moteur(0);
  Aller(0, 800);
  //aller sur puce 0, init temps
  Aller(0, -400);
  Aller(0, 400);
  temps_boucle_total = (uint32_t)millis();
  //rouler a fond
  Aller(0, 1000);
  temps_boucle_total = (uint32_t)millis() - temps_boucle_total ;
  //rouler cool pour lecture rfid precise
  Aller(0, -400);
  Vitesse_moteur(400); // on part pour un tour
  //  lire les puces, init du tableau, envoi /POS if(f) no_puce temps_en_ms (vitesse)
  //  des que puce 0 atteinte stocker temps_boucle_total nbre_de_puce
  // on positionnera les puces de facon regulière
  int i = 0;
  int puce = 0;
  for (puce = 1; puce <= nbr_de_puces; puce++) { // On doit déjà connaitre la puce 0
    unsigned char val;
    char tagValue[12];
    for (i = 0; i <= 11; i++) {
      if (rfid.read(val)) {
        tagValue[i] = '\0'; // Null
        // on saute avec le start 0x02 et stop bit 0x03
        if ((val == 0x0D) || (val == 0x0A) || (val == 0x03) || (val == 0x02)) {
          break;
        }
        Serial.print(i);
        Serial.print(" ");
        Serial.print((char)val);
        Serial.print(" ");
        Serial.println(val, HEX );
        tagValue[i] = (char)val;
        if ((i == 11) ) { // 11 digit read is complete
          tagValue[12] = '\0'; // Null
          //        Serial.print("Tag read: ");
          Serial.println(tagValue);
          RFID[puce] = tagValue ;
        }
      }
    }
  }
  commande = 0;
  Aller(0, -400); //retourner sur la puce 0 et y stopper
}
void Rendezvous(OSCMessage &msg) {       // A FINIR
  // /goto ii puce temps : se rendre a la puce xx en tant de temps
  commande = 3;
  Serial.println("/rdv");
  puce_a_trouver = (int) msg.getInt(0);
  if (no_puce_actuelle != puce_a_trouver) {
    temps_a_faire = (uint32_t)msg.getInt(1); // en s
    temps_depart = (uint32_t)(millis() * 1000);
    Vitesse_moteur((int)msg.getInt(1));
  }
}
void SeRendre(OSCMessage &msg) {       // A FINIR
  // /aller ii puce vitesse : se rendre a la puce xx a n vitesse
  commande = 4;
  Serial.println("/aller trouver");
  puce_a_trouver = (int)msg.getInt(0);
  Serial.println(puce_a_trouver);
  Serial.println(no_puce_actuelle);
  Serial.println(puce_derniere_lue);
  if (puce_derniere_lue != puce_a_trouver) {
    Vitesse_moteur((int)msg.getInt(1));
    Serial.println((int)Vitesse_moteur);
  }
}

void loop()
{
  // Lire la comande et executer le debut
  //OSCBundle msgIN;
  OSCMessage msgIN;
  int size;
  if ( (size = udpin.parsePacket()) > 0)
  {
    while (size--)
      msgIN.fill(udpin.read());

    if (!msgIN.hasError()) {
      msgIN.dispatch("/stop", ArretMoteur);           // stop le moteur immediatement commande=0
      msgIN.dispatch("/moteur", MoteurControl);  // met le moteur à la vitesse lue  commande=1
      //msgIN.route("/init", Initialisation); // fait un tour pour reperer les puces (puce0 fixe) commande=2
      //msgIN.route("/goto", Rendezvous);           // va à la puce dans le temps donné commande=3
      msgIN.dispatch("/aller", SeRendre);  // aller a la puce a une vitesse /aller ii puce vitesse commande=4
    }
    udpin.flush();
  }
  // Lire une puce et continuer une commande en cours
  if (commande != 2) {
    int puce_lue = Lit_rfid();
    if ((commande == 4 || commande == 3) && puce_lue == puce_a_trouver) {
      commande = 0;
      Serial.println("l trouve");
      Serial.println(puce_a_trouver);
      analogWrite(enA, 0);
      Message(puce_a_trouver);
      no_puce_actuelle = puce_a_trouver;
      puce_derniere_lue = puce_a_trouver;
      puce_a_trouver = 98;
    }

  }
  delay(100);
}
